# standard libraries
import collections
import contextlib
import copy
import logging
import math
import unittest

# third party libraries
import numpy

# local libraries
from nion.swift import Application
from nion.swift import DocumentController
from nion.swift.model import DataItem
from nion.swift.model import DocumentModel
from nion.swift.model import Graphics
from nion.ui import CanvasItem
from nion.ui import TestUI
from nion.utils import Geometry


class TestGraphicsClass(unittest.TestCase):

    def setUp(self):
        self.app = Application.Application(TestUI.UserInterface(), set_global=False)

    def tearDown(self):
        pass

    class Mapping(object):
        def __init__(self, size):
            self.size = size
        def map_point_image_norm_to_widget(self, p):
            return (p[0] * self.size[0], p[1] * self.size[1])
        def map_size_image_norm_to_widget(self, s):
            return (s[0] * self.size[0], s[1] * self.size[1])
        def map_point_widget_to_image_norm(self, p):
            return (p[0]/self.size[0], p[1]/self.size[1])

    def test_copy_graphic(self):
        rect_graphic = Graphics.RectangleGraphic()
        copy.deepcopy(rect_graphic)
        ellipse_graphic = Graphics.EllipseGraphic()
        copy.deepcopy(ellipse_graphic)
        line_graphic = Graphics.LineGraphic()
        copy.deepcopy(line_graphic)

    def test_rect_test(self):
        mapping = TestGraphicsClass.Mapping((1000, 1000))
        rect_graphic = Graphics.RectangleGraphic()
        rect_graphic.bounds = (0.25, 0.25), (0.5, 0.5)
        def get_font_metrics(font, text):
            FontMetrics = collections.namedtuple("FontMetrics", ["width", "height", "ascent", "descent", "leading"])
            return FontMetrics(width=(len(text) * 7.0), height=18, ascent=15, descent=3, leading=0)
        self.assertEqual(rect_graphic.test(mapping, get_font_metrics, (500, 500), move_only=False), ("all", False))
        self.assertEqual(rect_graphic.test(mapping, get_font_metrics, (250, 250), move_only=False)[0], "top-left")
        self.assertEqual(rect_graphic.test(mapping, get_font_metrics, (750, 750), move_only=False)[0], "bottom-right")
        self.assertEqual(rect_graphic.test(mapping, get_font_metrics, (250, 750), move_only=False)[0], "top-right")
        self.assertEqual(rect_graphic.test(mapping, get_font_metrics, (750, 250), move_only=False)[0], "bottom-left")
        self.assertEqual(rect_graphic.test(mapping, get_font_metrics, (250, 500), move_only=False), ("all", True))
        self.assertEqual(rect_graphic.test(mapping, get_font_metrics, (750, 500), move_only=False), ("all", True))
        self.assertEqual(rect_graphic.test(mapping, get_font_metrics, (500, 250), move_only=False), ("all", True))
        self.assertEqual(rect_graphic.test(mapping, get_font_metrics, (500, 750), move_only=False), ("all", True))
        self.assertIsNone(rect_graphic.test(mapping, get_font_metrics, (0, 0), move_only=False)[0])

    def test_line_test(self):
        mapping = TestGraphicsClass.Mapping((1000, 1000))
        line_graphic = Graphics.LineGraphic()
        line_graphic.start = (0.25,0.25)
        line_graphic.end = (0.75,0.75)
        def get_font_metrics(font, text):
            FontMetrics = collections.namedtuple("FontMetrics", ["width", "height", "ascent", "descent", "leading"])
            return FontMetrics(width=(len(text) * 7.0), height=18, ascent=15, descent=3, leading=0)
        self.assertEqual(line_graphic.test(mapping, get_font_metrics, (500, 500), move_only=True)[0], "all")
        self.assertEqual(line_graphic.test(mapping, get_font_metrics, (250, 250), move_only=True)[0], "start")
        self.assertEqual(line_graphic.test(mapping, get_font_metrics, (750, 750), move_only=True)[0], "end")
        self.assertEqual(line_graphic.test(mapping, get_font_metrics, (250, 250), move_only=False)[0], "start")
        self.assertEqual(line_graphic.test(mapping, get_font_metrics, (750, 750), move_only=False)[0], "end")
        self.assertIsNone(line_graphic.test(mapping, get_font_metrics, (240, 240), move_only=False)[0])
        self.assertIsNone(line_graphic.test(mapping, get_font_metrics, (760, 760), move_only=False)[0])
        self.assertIsNone(line_graphic.test(mapping, get_font_metrics, (0, 0), move_only=False)[0])

    def test_point_test(self):
        mapping = TestGraphicsClass.Mapping((1000, 1000))
        point_graphic = Graphics.PointGraphic()
        point_graphic.position = (0.25,0.25)
        def get_font_metrics(font, text):
            FontMetrics = collections.namedtuple("FontMetrics", ["width", "height", "ascent", "descent", "leading"])
            return FontMetrics(width=(len(text) * 7.0), height=18, ascent=15, descent=3, leading=0)
        self.assertEqual(point_graphic.test(mapping, get_font_metrics, (250, 250), move_only=True)[0], "all")
        self.assertEqual(point_graphic.test(mapping, get_font_metrics, (250 - 18, 250), move_only=True)[0], None)
        point_graphic.label = "Test"
        self.assertEqual(point_graphic.test(mapping, get_font_metrics, (250 - 18 - 6, 250), move_only=True)[0], "all")

    def test_spot_test(self):
        mapping = TestGraphicsClass.Mapping((1000, 1000))
        spot_graphic = Graphics.SpotGraphic()
        spot_graphic.bounds = (0.2, 0.2), (0.1, 0.1)

        def get_font_metrics(font, text):
            FontMetrics = collections.namedtuple("FontMetrics", ["width", "height", "ascent", "descent", "leading"])
            return FontMetrics(width=(len(text) * 7.0), height=18, ascent=15, descent=3, leading=0)

        self.assertEqual(spot_graphic.test(mapping, get_font_metrics, (200, 200), move_only=False), ("top-left", True))
        self.assertEqual(spot_graphic.test(mapping, get_font_metrics, (300, 200), move_only=False), ("bottom-left", True))
        self.assertEqual(spot_graphic.test(mapping, get_font_metrics, (300, 300), move_only=False), ("bottom-right", True))
        self.assertEqual(spot_graphic.test(mapping, get_font_metrics, (200, 300), move_only=False), ("top-right", True))
        self.assertEqual(spot_graphic.test(mapping, get_font_metrics, (800, 800), move_only=False), ("inverted-top-left", True))
        self.assertEqual(spot_graphic.test(mapping, get_font_metrics, (700, 800), move_only=False), ("inverted-bottom-left", True))
        self.assertEqual(spot_graphic.test(mapping, get_font_metrics, (700, 700), move_only=False), ("inverted-bottom-right", True))
        self.assertEqual(spot_graphic.test(mapping, get_font_metrics, (800, 700), move_only=False), ("inverted-top-right", True))
        self.assertIsNone(spot_graphic.test(mapping, get_font_metrics, (0, 0), move_only=False)[0])

    def test_create_all_graphic_by_dragging(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            t = [
                ("line", Graphics.LineGraphic),
                ("rectangle", Graphics.RectangleGraphic),
                ("ellipse", Graphics.EllipseGraphic),
                ("point", Graphics.PointGraphic),
                ("spot", Graphics.SpotGraphic),
                ("wedge", Graphics.WedgeGraphic),
                ("ring", Graphics.RingGraphic),
            ]
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            for tool_mode, graphic_type in t:
                document_controller.tool_mode = tool_mode
                self.assertEqual(0, len(display_item.graphics))
                display_panel.display_canvas_item.simulate_drag((500, 500), (600, 600))
                self.assertEqual(1, len(display_item.graphics))
                graphic = display_item.graphics[0]
                self.assertIsInstance(graphic, graphic_type)
                display_item.remove_graphic(graphic)

    def test_spot_mask_is_sensible_when_smaller_than_one_pixel(self):
        spot_graphic = Graphics.SpotGraphic()
        spot_graphic.bounds = (0.5, 0.5), (0.02, 0.02)
        mask_data = spot_graphic.get_mask((10, 10))
        self.assertEqual(mask_data.shape, (10, 10))
        self.assertTrue(numpy.array_equal(mask_data, numpy.zeros((10, 10))))

    def test_spot_mask_is_sensible_when_outside_bounds(self):
        spot_graphic = Graphics.SpotGraphic()
        spot_graphic.bounds = (1.5, 1.5), (0.1, 0.1)
        mask_data = spot_graphic.get_mask((10, 10))
        self.assertEqual(mask_data.shape, (10, 10))
        self.assertTrue(numpy.array_equal(mask_data, numpy.zeros((10, 10))))
        mask_data = spot_graphic.get_mask((10, 10))
        spot_graphic.bounds = (-0.5, -0.5), (0.1, 0.1)
        self.assertEqual(mask_data.shape, (10, 10))
        self.assertTrue(numpy.array_equal(mask_data, numpy.zeros((10, 10))))

    def test_spot_mask_is_sensible_when_partially_outside_bounds(self):
        spot_graphic = Graphics.SpotGraphic()
        spot_graphic.bounds = (0.75, 0.75), (0.5, 0.5)
        mask_data = spot_graphic.get_mask((10, 10))
        self.assertEqual(mask_data.shape, (10, 10))
        self.assertFalse(numpy.array_equal(mask_data, numpy.zeros((10, 10))))
        spot_graphic.bounds = (-0.25, -0.25), (0.5, 0.5)
        mask_data = spot_graphic.get_mask((10, 10))
        self.assertEqual(mask_data.shape, (10, 10))
        self.assertFalse(numpy.array_equal(mask_data, numpy.zeros((10, 10))))

    def assertAlmostEqualPoint(self, p1, p2, e=0.00001):
        if not(Geometry.distance(p1, p2) < e):
            logging.debug("%s != %s", p1, p2)
        self.assertTrue(Geometry.distance(p1, p2) < e)

    def assertAlmostEqualSize(self, s1, s2, e=0.00001):
        if not(abs(s2.height - s1.height) < e) or not(abs(s2.width - s1.width) < e):
            logging.debug("%s != %s", s1, s2)
        self.assertTrue(abs(s2.height - s1.height) < e and abs(s2.width - s1.width) < e)

    def assertAlmostEqualRect(self, r1, r2, e=0.00001):
        if not(Geometry.distance(r1[0], r2[0]) < e and Geometry.distance(r1[1], r2[1]) < e):
            logging.debug("%s != %s", r1, r2)
        self.assertTrue(Geometry.distance(r1[0], r2[0]) < e and Geometry.distance(r1[1], r2[1]) < e)

    def test_dragging_regions(self):
        # make the document controller
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        display_panel = document_controller.selected_display_panel
        data_item = DataItem.DataItem(numpy.zeros((10, 10)))
        document_model.append_data_item(data_item)
        display_item = document_model.get_display_item_for_data_item(data_item)
        display_panel.set_display_panel_display_item(display_item)
        header_height = display_panel.header_canvas_item.header_height
        display_panel.root_container.layout_immediate((1000 + header_height, 1000))

        def get_extended_attr(object, extended_name):
            initial_value = object
            for sub_property in extended_name.split("."):
                initial_value = getattr(initial_value, sub_property)
            return initial_value

        def map_string(s, m1, m2):
            for k, v in m1.items():
                s = s.replace(k, v)
            for k, v in m2.items():
                s = s.replace(k, v)
            return s

        class ScalarCoordinate(object):

            def __init__(self, v_value=None, h_value=None):
                self.v_value = v_value
                self.h_value = h_value

            @property
            def value(self):
                return self.v_value if self.v_value is not None else self.h_value

        def reflect_rect(r, v, h):
            if v and not h:
                return Geometry.FloatRect(Geometry.FloatPoint(1.0 - r.bottom, r.left), r.size)
            elif v and h:
                return Geometry.FloatRect(Geometry.FloatPoint(1.0 - r.bottom, 1.0 - r.right), r.size)
            elif not v and h:
                return Geometry.FloatRect(Geometry.FloatPoint(r.top, 1.0 - r.right), r.size)
            else:
                return r

        def reflect_point(p, v, h):
            y = 1.0 - p.y if v else p.y
            x = 1.0 - p.x if h else p.x
            return Geometry.FloatPoint(y, x)

        def reflect_size(s, v, h):
            height = 1.0 - s.height if v else s.height
            width = 1.0 - s.width if h else s.width
            return Geometry.FloatSize(height, width)

        def reflect_scalar(s, v, h):
            v_value = 1.0 - s.v_value if s.v_value is not None and v else s.v_value
            h_value = 1.0 - s.h_value if s.h_value is not None and h else s.h_value
            return ScalarCoordinate(v_value, h_value)

        def reflect_value(value, v, h, mapping_v1, mapping_v2, mapping_h1, mapping_h2):
            if isinstance(value, Geometry.FloatRect):
                return reflect_rect(value, v, h)
            elif isinstance(value, Geometry.FloatPoint):
                return reflect_point(value, v, h)
            elif isinstance(value, Geometry.FloatSize):
                return reflect_size(value, v, h)
            elif isinstance(value, ScalarCoordinate):
                return reflect_scalar(value, v, h)
            elif isinstance(value, str):
                if v:
                    value = map_string(value, mapping_v1, mapping_v2)
                if h:
                    value = map_string(value, mapping_h1, mapping_h2)
                return value
            elif isinstance(value, float):
                return value
            raise Exception("Unknown value type %s", type(value))
            return None

        def reflect(d, v, h):
            d = copy.deepcopy(d)
            mapping_v1 = {"top": "was_top", "bottom": "was_bottom"}
            mapping_v2 = {"was_top": "bottom", "was_bottom": "top"}
            mapping_h1 = {"left": "was_left", "right": "was_right"}
            mapping_h2 = {"was_left": "right", "was_right": "left"}
            # name
            if v:
                d["name"] = map_string(d["name"], mapping_v1, mapping_v2)
            if h:
                d["name"] = map_string(d["name"], mapping_h1, mapping_h2)
            # input
            old_input = d["input"]["properties"]
            new_input = dict()
            for property, expected_value in old_input.items():
                if v:
                    property = map_string(property, mapping_v1, mapping_v2)
                if h:
                    property = map_string(property, mapping_h1, mapping_h2)
                new_input[property] = reflect_value(expected_value, v, h, mapping_v1, mapping_v2, mapping_h1, mapping_h2)
            d["input"]["properties"] = new_input
            # drag
            drag = d["drag"]
            c0, c1 = drag[0:2]
            if v:
                c0 = 1000 - c0[0], c0[1]
                c1 = 1000 - c1[0], c1[1]
            if h:
                c0 = c0[0], 1000 - c0[1]
                c1 = c1[0], 1000 - c1[1]
            new_drag = list((c0, c1))
            new_drag.extend(drag[2:])
            d["drag"] = tuple(new_drag)
            # output
            old_output = d["output"]["properties"]
            new_output = dict()
            for property, expected_value in old_output.items():
                if v:
                    property = map_string(property, mapping_v1, mapping_v2)
                if h:
                    property = map_string(property, mapping_h1, mapping_h2)
                new_output[property] = reflect_value(expected_value, v, h, mapping_v1, mapping_v2, mapping_h1, mapping_h2)
            d["output"]["properties"] = new_output
            return d

        def do_drag_test(d, e=0.00001):
            # logging.debug("test %s", d["name"])
            region = Graphics.factory(lambda t: d["input"]["type"])
            for property, initial_value in d["input"]["properties"].items():
                setattr(region, property, initial_value)
            initial_values = dict()
            for property, expected_value in d["output"]["properties"].items():
                if isinstance(expected_value, str):
                    initial_values[expected_value] = get_extended_attr(region, expected_value)
            for constraint in d["input"].get("constraints", list()):
                if constraint == "bounds":
                    region.is_bounds_constrained = True
                elif constraint == "shape":
                    region.is_shape_locked = True
                elif constraint == "position":
                    region.is_position_locked = True
            display_item.add_graphic(region)
            display_item.graphic_selection.set(0)
            display_panel.display_canvas_item.simulate_drag(*d["drag"])
            for property, expected_value in d["output"]["properties"].items():
                actual_value = get_extended_attr(region, property)
                # logging.debug("%s: %s == %s ?", property, actual_value, expected_value)
                if isinstance(expected_value, str):
                    if isinstance(actual_value, Geometry.FloatRect):
                        self.assertAlmostEqualRect(actual_value, initial_values[expected_value], e)
                    elif isinstance(actual_value, Geometry.FloatPoint):
                        self.assertAlmostEqualPoint(actual_value, initial_values[expected_value], e)
                    elif isinstance(actual_value, Geometry.FloatSize):
                        self.assertAlmostEqualSize(actual_value, initial_values[expected_value], e)
                    elif isinstance(actual_value, ScalarCoordinate):
                        self.assertAlmostEqual(actual_value.value, initial_values[expected_value])
                    else:
                        raise Exception("Unknown value type %s", type(actual_value))
                else:
                    if isinstance(actual_value, Geometry.FloatRect):
                        self.assertAlmostEqualRect(actual_value, expected_value, e)
                    elif isinstance(actual_value, Geometry.FloatPoint):
                        self.assertAlmostEqualPoint(actual_value, expected_value, e)
                    elif isinstance(actual_value, Geometry.FloatSize):
                        self.assertAlmostEqualSize(actual_value, expected_value, e)
                    elif isinstance(actual_value, float):
                        self.assertAlmostEqual(actual_value, expected_value.value)
                    else:
                        raise Exception("Unknown value type %s", type(actual_value))
            display_item.remove_graphic(region)
            region.is_bounds_constrained = False

        def rotate(p, o, angle):
            origin = Geometry.FloatPoint.make(o)
            point = Geometry.FloatPoint.make(p)
            delta = point - origin
            angle_sin = math.sin(angle)
            angle_cos = math.cos(angle)
            return origin + Geometry.FloatPoint(x=delta.x * angle_cos + delta.y * angle_sin, y=delta.y * angle_cos - delta.x * angle_sin)

        def round1000(p):
            return Geometry.FloatPoint(y=int(p.y * 1000) / 1000, x=int(p.x * 1000) / 1000)

        # rectangle top-left

        d = {
            "name": "drag top-left corner outside of bounds with no constraints",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect(origin=(0.21, 0.22), size=(0.31, 0.32)) }
            },
            "drag": [(210, 220), (-190, -180)],
            "output": {
                "properties": {
                    "_bounds.top_left": Geometry.FloatPoint(y=-0.19, x=-0.18),
                    "_bounds.bottom_right": "_bounds.bottom_right"
                },
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        for rotation in (2 * math.pi / 8, -2 * math.pi / 8):

            d = {
                "name": "drag rotated top-left corner outside of bounds with no constraints",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect(origin=(0.4, 0.3), size=(0.2, 0.4)), "rotation": rotation }
                },
                "drag": [rotate((400, 300), (500, 500), rotation), rotate((450, 250), (500, 500), rotation)],
                "output": {
                    "properties": {
                        "_rotated_top_left": rotate((0.45, 0.25), (0.5, 0.5), rotation),
                        "_rotated_bottom_right": "_rotated_bottom_right"
                    },
                }
            }

            # rotation odd reflections are not valid
            do_drag_test(reflect(d, False, False), e=0.001)
            do_drag_test(reflect(d, True, True), e=0.001)

            d = {
                "name": "drag rotated top-right corner outside of bounds with no constraints",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect(origin=(0.4, 0.3), size=(0.2, 0.4)), "rotation": rotation }
                },
                "drag": [rotate((400, 700), (500, 500), rotation), rotate((450, 750), (500, 500), rotation)],
                "output": {
                    "properties": {
                        "_rotated_top_right": rotate((0.45, 0.75), (0.5, 0.5), rotation),
                        "_rotated_bottom_left": "_rotated_bottom_left"
                    },
                }
            }

            # rotation odd reflections are not valid
            do_drag_test(reflect(d, False, False), e=0.001)
            do_drag_test(reflect(d, True, True), e=0.001)

            d = {
                "name": "drag rotated top-left corner outside of bounds from center",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect(origin=(0.4, 0.3), size=(0.2, 0.4)), "rotation": rotation }
                },
                "drag": [rotate((400, 300), (500, 500), rotation), rotate((450, 250), (500, 500), rotation), CanvasItem.KeyboardModifiers(alt=True)],
                "output": {
                    "properties": {
                        "_rotated_top_left": rotate((0.45, 0.25), (0.5, 0.5), rotation),
                        "_bounds.center": "_bounds.center"
                    },
                }
            }

            # rotation odd reflections are not valid
            do_drag_test(reflect(d, False, False), e=0.001)
            do_drag_test(reflect(d, True, True), e=0.001)

            d = {
                "name": "drag rotated top-right corner outside of bounds from center",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect(origin=(0.4, 0.3), size=(0.2, 0.4)), "rotation": rotation }
                },
                "drag": [rotate((400, 700), (500, 500), rotation), rotate((450, 750), (500, 500), rotation), CanvasItem.KeyboardModifiers(alt=True)],
                "output": {
                    "properties": {
                        "_rotated_top_right": rotate((0.45, 0.75), (0.5, 0.5), rotation),
                        "_bounds.center": "_bounds.center"
                    },
                }
            }

            # rotation odd reflections are not valid
            do_drag_test(reflect(d, False, False), e=0.001)
            do_drag_test(reflect(d, True, True), e=0.001)

        d = {
            "name": "drag top-left corner outside of bounds with bounds constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.21, 0.22), (0.31, 0.32)) },
                "constraints": ["bounds"]
            },
            "drag": [(210, 220), (-190, -180)],
            "output": {
                "properties": {
                    "_bounds.top_left": Geometry.FloatPoint(),
                    "_bounds.bottom_right": "_bounds.bottom_right"
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag top-left corner from center towards top left with bounds constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.25, 0.25), (0.5, 0.5)) },
                "constraints": ["bounds"]
            },
            "drag": [(250, 250), (150, 140), CanvasItem.KeyboardModifiers(alt=True)],
            "output": {
                "properties": {
                    "_bounds.center": "_bounds.center",
                    "_bounds.top_left": Geometry.FloatPoint(0.15, 0.14),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag top-left corner from center towards top left with position constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.25, 0.25), (0.5, 0.5)) },
                "constraints": ["position"]
            },
            "drag": [(250, 250), (150, 140), CanvasItem.KeyboardModifiers()],
            "output": {
                "properties": {
                    "_bounds.center": "_bounds.center",
                    "_bounds.top_left": Geometry.FloatPoint(0.15, 0.14),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag top-left corner from center outside of bounds to the top left with bounds constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.21, 0.22), (0.31, 0.32)) },
                "constraints": ["bounds"]
            },
            "drag": [(210, 220), (-190, -180), CanvasItem.KeyboardModifiers(alt=True)],
            "output": {
                "properties": {
                    "_bounds.center": "_bounds.center",
                    "_bounds.top_left": Geometry.FloatPoint(),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag top-left corner from center outside of bounds to the bottom with bounds constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.21, 0.22), (0.31, 0.32)) },
                "constraints": ["bounds"]
            },
            "drag": [(210, 220), (1500, 220), CanvasItem.KeyboardModifiers(alt=True)],
            "output": {
                "properties": {
                    "_bounds.center": "_bounds.center",
                    "_bounds.top": ScalarCoordinate(v_value=0.0),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag top-left corner from center outside of bounds to the right with bounds constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.21, 0.22), (0.31, 0.32)) },
                "constraints": ["bounds"]
            },
            "drag": [(210, 220), (210, 1500), CanvasItem.KeyboardModifiers(alt=True)],
            "output": {
                "properties": {
                    "_bounds.center": "_bounds.center",
                    "_bounds.left": ScalarCoordinate(h_value=0.0),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag squared top-left corner from outside of bounds to the bottom right with bounds constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.2, 0.3), (0.4, 0.5)) },
                "constraints": ["bounds"]
            },
            "drag": [(200, 300), (1500, 1500), CanvasItem.KeyboardModifiers(shift=True)],
            "output": {
                "properties": {
                    "_bounds.top_left": "_bounds.bottom_right",
                    "_bounds.bottom": ScalarCoordinate(v_value=0.8),
                    "_bounds.right": ScalarCoordinate(h_value=1.0),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag squared top-left corner from in one direction",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.4, 0.4), (0.2, 0.2)) },  # center 0.5, 0.5
                "constraints": ["bounds"]
            },
            "drag": [(400, 400), (200, 400), CanvasItem.KeyboardModifiers(shift=True)],
            "output": {
                "properties": {
                    "_bounds.bottom_right": "_bounds.bottom_right",
                    "_bounds.top_left": Geometry.FloatPoint(0.2, 0.2),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag squared top-left corner from outside of bounds to the bottom right with bounds constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.1, 0.3), (0.2, 0.2)) },  # center 0.2, 0.4
                "constraints": ["bounds"]
            },
            "drag": [(100, 300), (-200, -200), CanvasItem.KeyboardModifiers(shift=True, alt=True)],
            "output": {
                "properties": {
                    "_bounds.center": "_bounds.center",
                    "_bounds.left": ScalarCoordinate(h_value=0.2),
                    "_bounds.top": ScalarCoordinate(v_value=0.0),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag squared top-left corner with bounds, center, square constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.4, 0.4), (0.2, 0.2)) },  # center 0.5, 0.5
                "constraints": ["bounds"]
            },
            "drag": [(400, 400), (300, 200), CanvasItem.KeyboardModifiers(shift=True, alt=True)],
            "output": {
                "properties": {
                    "_bounds.center": "_bounds.center",
                    "_bounds.top_left": Geometry.FloatPoint(0.3, 0.3),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag squared top-left corner with bounds, center, square constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.7, 0.7), (0.2, 0.2)) },  # center 0.8, 0.8
                "constraints": ["bounds"]
            },
            "drag": [(700, 700), (0, 0), CanvasItem.KeyboardModifiers(shift=True, alt=True)],
            "output": {
                "properties": {
                    "_bounds.center": "_bounds.center",
                    "_bounds.right": ScalarCoordinate(h_value=1.0),
                    "_bounds.bottom": ScalarCoordinate(v_value=1.0),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag top-left to top-left with shape constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.3, 0.2), (0.2, 0.2)) },  # center 0.4, 0.3
                "constraints": ["shape"]
            },
            "drag": [(300, 200), (-100, -100), CanvasItem.KeyboardModifiers()],
            "output": {
                "properties": {
                    "_bounds.center": Geometry.FloatPoint(),
                    "_bounds.size": "_bounds.size",
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag all to top-left with no constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.3, 0.2), (0.2, 0.2)) },  # center 0.4, 0.3
            },
            "drag": [(400, 300), (0, 0), CanvasItem.KeyboardModifiers()],
            "output": {
                "properties": {
                    "_bounds.center": Geometry.FloatPoint(),
                    "_bounds.size": "_bounds.size",
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag all top-left with bounds constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.3, 0.2), (0.2, 0.2)) },  # center 0.4, 0.3
                "constraints": ["bounds"]
            },
            "drag": [(400, 300), (0, 0), CanvasItem.KeyboardModifiers()],
            "output": {
                "properties": {
                    "_bounds.center": Geometry.FloatPoint(0.1, 0.1),
                    "_bounds.size": "_bounds.size",
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag all bottom-right with bounds constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.3, 0.2), (0.2, 0.2)) },  # center 0.4, 0.3
                "constraints": ["bounds"]
            },
            "drag": [(400, 300), (1000, 1000), CanvasItem.KeyboardModifiers()],
            "output": {
                "properties": {
                    "_bounds.center": Geometry.FloatPoint(0.9, 0.9),
                    "_bounds.size": "_bounds.size",
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "drag all with restrict constraint",
            "input": {
                "type": "rect-graphic",
                "properties": { "_bounds": Geometry.FloatRect((0.4, 0.4), (0.2, 0.2)) },  # center 0.5, 0.5
            },
            "drag": [(500, 500), (800, 600), CanvasItem.KeyboardModifiers(shift=True)],
            "output": {
                "properties": {
                    "_bounds.center": Geometry.FloatPoint(0.8, 0.5),
                    "_bounds.size": "_bounds.size",
                }
            }
        }

        # vertical reflections not valid
        for h in (False, True):
            do_drag_test(reflect(d, False, h))

        # point

        d = {
            "name": "point drag with no constraint",
            "input": {
                "type": "point-graphic",
                "properties": { "_position": Geometry.FloatPoint(0.2, 0.3) },
                # "constraints": ["bounds"]
            },
            "drag": [(200, 300), (-100, -100), CanvasItem.KeyboardModifiers()],
            "output": {
                "properties": {
                    "_position": Geometry.FloatPoint(-0.1, -0.1),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "point drag with bounds constraint",
            "input": {
                "type": "point-graphic",
                "properties": { "_position": Geometry.FloatPoint(0.2, 0.3) },
                "constraints": ["bounds"]
            },
            "drag": [(200, 300), (-100, -100), CanvasItem.KeyboardModifiers()],
            "output": {
                "properties": {
                    "_position": Geometry.FloatPoint(),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "point drag with restrict",
            "input": {
                "type": "point-graphic",
                "properties": { "_position": Geometry.FloatPoint(0.2, 0.3) },
                "constraints": ["bounds"]
            },
            "drag": [(200, 300), (100, 100), CanvasItem.KeyboardModifiers(shift=True)],
            "output": {
                "properties": {
                    "_position": Geometry.FloatPoint(0.2, 0.1),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        # line regions

        d = {
            "name": "line drag with no constraints",
            "input": {
                "type": "line-graphic",
                "properties": {
                    "_start": Geometry.FloatPoint(0.2, 0.3),
                    "_end": Geometry.FloatPoint(0.6, 0.5),
                },
                # "constraints": ["bounds"]
            },
            "drag": [(400, 400), (600, 700), CanvasItem.KeyboardModifiers()],
            "output": {
                "properties": {
                    "_start": Geometry.FloatPoint(0.4, 0.6),
                    "_end": Geometry.FloatPoint(0.8, 0.8),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "line start drag with restrict",
            "input": {
                "type": "line-graphic",
                "properties": {
                    "_start": Geometry.FloatPoint(0.2, 0.3),
                    "_end": Geometry.FloatPoint(0.6, 0.5),
                },
                "constraints": ["shape"]
            },
            "drag": [(200, 300), (400, 600), CanvasItem.KeyboardModifiers()],
            "output": {
                "properties": {
                    "_start": Geometry.FloatPoint(0.4, 0.6),
                    "_end": Geometry.FloatPoint(0.8, 0.8),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "line drag with bounds constraint",
            "input": {
                "type": "line-graphic",
                "properties": {
                    "_start": Geometry.FloatPoint(0.2, 0.3),
                    "_end": Geometry.FloatPoint(0.6, 0.5),
                },
                "constraints": ["bounds"]
            },
            "drag": [(400, 400), (1000, 1000), CanvasItem.KeyboardModifiers()],
            "output": {
                "properties": {
                    "_start": Geometry.FloatPoint(0.6, 0.8),
                    "_end": Geometry.FloatPoint(1.0, 1.0),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "line start drag with no constraints",
            "input": {
                "type": "line-graphic",
                "properties": {
                    "_start": Geometry.FloatPoint(0.2, 0.3),
                    "_end": Geometry.FloatPoint(0.6, 0.5),
                },
                # "constraints": ["bounds"]
            },
            "drag": [(200, 300), (600, 700), CanvasItem.KeyboardModifiers()],
            "output": {
                "properties": {
                    "_start": Geometry.FloatPoint(0.6, 0.7),
                    "_end": Geometry.FloatPoint(0.6, 0.5),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "line start drag with bounds constraint",
            "input": {
                "type": "line-graphic",
                "properties": {
                    "_start": Geometry.FloatPoint(0.2, 0.3),
                    "_end": Geometry.FloatPoint(0.6, 0.5),
                },
                "constraints": ["bounds"]
            },
            "drag": [(200, 300), (-100, 300), CanvasItem.KeyboardModifiers()],
            "output": {
                "properties": {
                    "_start": Geometry.FloatPoint(0.0, 0.3),
                    "_end": Geometry.FloatPoint(0.6, 0.5),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        d = {
            "name": "line start drag with shape constraint",
            "input": {
                "type": "line-graphic",
                "properties": {
                    "_start": Geometry.FloatPoint(0.2, 0.3),
                    "_end": Geometry.FloatPoint(0.6, 0.5),
                },
                "constraints": ["bounds"]
            },
            "drag": [(200, 300), (500, 500), CanvasItem.KeyboardModifiers(shift=True)],
            "output": {
                "properties": {
                    "_start": Geometry.FloatPoint(0.5, 0.5),
                    "_end": Geometry.FloatPoint(0.6, 0.5),
                }
            }
        }

        for v in (False, True):
            for h in (False, True):
                do_drag_test(reflect(d, v, h))

        document_controller.close()

    def test_selected_graphics_get_priority_when_dragging_middles(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            rect_graphic1 = Graphics.RectangleGraphic()
            rect_graphic2 = Graphics.RectangleGraphic()
            rect_graphic1.bounds = (0.4, 0.4), (0.2, 0.2)
            rect_graphic2.bounds = (0.25, 0.25), (0.5, 0.5)
            display_item.add_graphic(rect_graphic1)
            display_item.add_graphic(rect_graphic2)
            display_item.graphic_selection.set(0)
            # now the smaller rectangle (selected) is behind the larger one
            display_panel.display_canvas_item.simulate_drag((500, 500), (600, 600))
            # make sure the smaller one gets dragged
            self.assertAlmostEqualPoint(Geometry.FloatRect.make(rect_graphic1.bounds).center, Geometry.FloatPoint(y=0.6, x=0.6))
            self.assertAlmostEqualPoint(Geometry.FloatRect.make(rect_graphic2.bounds).center, Geometry.FloatPoint(y=0.5, x=0.5))

    def test_removing_graphic_from_display_closes_it(self):
        # make the document controller
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            graphic = Graphics.PointGraphic()
            display_item.add_graphic(graphic)
            self.assertFalse(graphic._closed)
            display_item.remove_graphic(graphic)
            self.assertTrue(graphic._closed)

    def test_removing_data_item_closes_graphic_attached_to_display(self):
        # make the document controller
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            graphic = Graphics.PointGraphic()
            display_item.add_graphic(graphic)
            self.assertFalse(graphic._closed)
            document_model.remove_data_item(data_item)
            self.assertTrue(graphic._closed)

    def test_removing_region_closes_associated_drawn_graphic(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            point_region = Graphics.PointGraphic()
            display_item.add_graphic(point_region)
            drawn_graphic = display_item.graphics[0]
            self.assertFalse(drawn_graphic._closed)
            display_item.remove_graphic(point_region)
            self.assertTrue(drawn_graphic._closed)

    def test_removing_data_item_closes_associated_drawn_graphic(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            point_region = Graphics.PointGraphic()
            display_item.add_graphic(point_region)
            drawn_graphic = display_item.graphics[0]
            self.assertFalse(drawn_graphic._closed)
            document_model.remove_data_item(data_item)
            self.assertTrue(drawn_graphic._closed)

    def test_removing_graphic_with_dependent_data_removes_dependent_data(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            crop_region = Graphics.RectangleGraphic()
            display_item.add_graphic(crop_region)
            cropped_data_item = document_model.get_crop_new(display_item, crop_region)
            self.assertIn(cropped_data_item, document_model.data_items)
            display_item.remove_graphic(crop_region)
            self.assertNotIn(cropped_data_item, document_model.data_items)

    def test_removing_one_of_two_graphics_with_dependent_data_only_removes_the_one(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            crop_region1 = Graphics.RectangleGraphic()
            display_item.add_graphic(crop_region1)
            crop_region2 = Graphics.RectangleGraphic()
            display_item.add_graphic(crop_region2)
            document_model.get_crop_new(display_item, crop_region1)
            document_model.get_crop_new(display_item, crop_region2)
            self.assertIn(crop_region1, display_item.graphics)
            self.assertIn(crop_region2, display_item.graphics)
            display_item.remove_graphic(crop_region1)
            self.assertIn(crop_region2, display_item.graphics)

    def test_changing_data_length_does_not_update_graphics(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((100, ), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            interval_graphic = Graphics.IntervalGraphic()
            interval_graphic.interval = (0.25, 0.75)
            display_item.add_graphic(interval_graphic)
            self.assertEqual(display_item.displayed_dimensional_scales[-1], 100)
            data_item.set_data(numpy.zeros((50, ), numpy.uint32))
            self.assertEqual(interval_graphic.interval, (0.25, 0.75))

    def test_changing_data_scale_does_not_update_graphics(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((100, ), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            interval_graphic = Graphics.IntervalGraphic()
            interval_graphic.interval = (0.25, 0.75)
            display_item.add_graphic(interval_graphic)
            self.assertEqual(display_item.displayed_dimensional_scales[-1], 100)
            display_item.dimensional_scales = (1,)
            self.assertEqual(interval_graphic.interval, (0.25, 0.75))

if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



# standard libraries
import threading

# third party libraries
# None

# local libraries
from nion.swift.model import Utility
from nion.ui import CanvasItem
from nion.ui import DrawingContext
from nion.utils import Geometry


class DisplayScriptCanvasItemDelegate:
    # interface must be implemented by the delegate

    def begin_mouse_tracking(self) -> None: ...

    def end_mouse_tracking(self, undo_command) -> None: ...

    def delete_key_pressed(self) -> None: ...

    def enter_key_pressed(self) -> None: ...

    def cursor_changed(self, pos): ...

    def show_display_context_menu(self, gx, gy) -> bool: ...

    @property
    def tool_mode(self) -> str: return str()


class DisplayScriptCanvasItem(CanvasItem.LayerCanvasItem):
    """Display a custom display using a script.

    Callers are expected to pass in a font metrics function and a delegate.
    """

    def __init__(self, get_font_metrics_fn, delegate, event_loop, draw_background: bool=True):
        super().__init__()

        self.__get_font_metrics_fn = get_font_metrics_fn
        self.delegate = delegate

        self.__drawing_context_lock = threading.RLock()
        self.__drawing_context = DrawingContext.DrawingContext()

        self.__display_data = None
        self.__display_script = None

        self.__closing_lock = threading.RLock()
        self.__closed = False

        self.__data = None
        self.__last_data = None

        # canvas items get added back to front
        # create the child canvas items
        # the background
        self.add_canvas_item(CanvasItem.BackgroundCanvasItem())

        # frame rate
        self.__display_frame_rate_id = None
        self.__display_frame_rate_last_index = 0

    def close(self):
        # call super
        with self.__closing_lock:
            self.__closed = True
        super().close()

    @property
    def default_aspect_ratio(self):
        return 1.0

    def update_display_values(self, display_values_list) -> None:
        self.__display_data = display_values_list[0].data_and_metadata if display_values_list else None

    def update_display_properties(self, display_calibration_info, display_properties, display_layers) -> None:
        self.__display_script = display_properties.get("display_script")
        self.update()

    def update_graphics(self, graphics, graphic_selection, display_calibration_info):
        pass

    def handle_auto_display(self) -> bool:
        # enter key has been pressed
        return False

    def _inserted(self, container):
        # make sure we get 'prepare_render' calls
        self.register_prepare_canvas_item(self)

    def _removed(self, container):
        # turn off 'prepare_render' calls
        self.unregister_prepare_canvas_item(self)

    def prepare_render(self):
        data_and_metadata = self.__display_data
        display_script = self.__display_script
        if data_and_metadata:
            # this method may trigger a layout of its parent scroll area. however, the parent scroll
            # area may already be closed. this is a stop-gap guess at a solution - the basic idea being
            # that this object is not closeable while this method is running; and this method should not
            # run if the object is already closed.
            with self.__closing_lock:
                if self.__closed:
                    return
                assert not self.__closed
                # Update the display state.
                rect = self.canvas_bounds
                if rect is not None:
                    g = dict()
                    drawing_context = DrawingContext.DrawingContext()
                    g["drawing_context"] = drawing_context
                    g["display_data_and_metadata"] = data_and_metadata
                    g["bounds"] = rect
                    g["get_font_metrics_fn"] = self.__get_font_metrics_fn
                    l = dict()
                    try:
                        # print(code)
                        compiled = compile(display_script, "expr", "exec")
                        exec(compiled, g, l)
                    except Exception as e:
                        # import sys, traceback
                        # traceback.print_exc()
                        # traceback.format_exception(*sys.exc_info())
                        print(str(e) or "Unable to evaluate display script.")  # a stack trace would be too much information right now

                    with self.__drawing_context_lock:
                        self.__drawing_context = drawing_context

    def _update_self_layout(self, canvas_origin, canvas_size, *, immediate=False):
        super()._update_self_layout(canvas_origin, canvas_size, immediate=immediate)
        self.prepare_render()

    def _repaint(self, drawing_context):
        super()._repaint(drawing_context)

        with self.__drawing_context_lock:
            drawing_context.add(self.__drawing_context)

        if self.__display_frame_rate_id:
            Utility.fps_tick("display_"+self.__display_frame_rate_id)

        if self.__display_frame_rate_id:
            fps = Utility.fps_get("display_"+self.__display_frame_rate_id)
            fps2 = Utility.fps_get("frame_"+self.__display_frame_rate_id)
            fps3 = Utility.fps_get("update_"+self.__display_frame_rate_id)

            rect = self.canvas_bounds

            with drawing_context.saver():
                font = "normal 11px serif"
                text_pos = Geometry.IntPoint(y=rect[0][0], x=rect[0][1] + rect[1][1] - 100)
                drawing_context.begin_path()
                drawing_context.move_to(text_pos.x, text_pos.y)
                drawing_context.line_to(text_pos.x + 120, text_pos.y)
                drawing_context.line_to(text_pos.x + 120, text_pos.y + 60)
                drawing_context.line_to(text_pos.x, text_pos.y + 60)
                drawing_context.close_path()
                drawing_context.fill_style = "rgba(255, 255, 255, 0.6)"
                drawing_context.fill()
                drawing_context.font = font
                drawing_context.text_baseline = "middle"
                drawing_context.text_align = "left"
                drawing_context.fill_style = "#000"
                drawing_context.fill_text("display:" + fps, text_pos.x + 8, text_pos.y + 10)
                drawing_context.fill_text("frame:" + fps2, text_pos.x + 8, text_pos.y + 30)
                drawing_context.fill_text("update:" + fps3, text_pos.x + 8, text_pos.y + 50)

    def mouse_entered(self):
        if super().mouse_entered():
            return True
        self.__mouse_in = True
        return True

    def mouse_exited(self):
        if super().mouse_exited():
            return True
        self.__mouse_in = False
        self.__update_cursor_info()
        if self.delegate:  # allow display to work without delegate
            # whenever the cursor exits, clear the cursor display
            self.delegate.cursor_changed(None)
        return True

    def mouse_double_clicked(self, x, y, modifiers):
        if super().mouse_clicked(x, y, modifiers):
            return True
        if self.delegate.tool_mode == "pointer":
            pass  # pos = Geometry.IntPoint(x=x, y=y)
        return False

    def mouse_position_changed(self, x, y, modifiers):
        if super().mouse_position_changed(x, y, modifiers):
            return True
        if self.delegate.tool_mode == "pointer":
            self.cursor_shape = "arrow"
        self.__last_mouse = Geometry.IntPoint(x=x, y=y)
        self.__update_cursor_info()
        return True

    def mouse_pressed(self, x, y, modifiers):
        if super().mouse_pressed(x, y, modifiers):
            return True
        self.delegate.begin_mouse_tracking()
        if self.delegate.tool_mode == "pointer":
            pass
        return False

    def mouse_released(self, x, y, modifiers):
        if super().mouse_released(x, y, modifiers):
            return True
        return False

    def context_menu_event(self, x, y, gx, gy):
        return self.delegate.show_display_context_menu(gx, gy)

    # ths message comes from the widget
    def key_pressed(self, key):
        if super().key_pressed(key):
            return True
        # only handle keys if we're directly embedded in an image panel
        if key.is_delete:
            self.delegate.delete_key_pressed()
            return True
        if key.is_enter_or_return:
            self.delegate.enter_key_pressed()
            return True
        if key.key == 70 and key.modifiers.shift and key.modifiers.alt:
            if self.__display_frame_rate_id is None:
                self.__display_frame_rate_id = str(id(self))
            else:
                self.__display_frame_rate_id = None
            return True
        return False

    def __update_cursor_info(self):
        if not self.delegate:  # allow display to work without delegate
            return
        if self.__mouse_in and self.__last_mouse:
            pass  # self.delegate.cursor_changed(pos_1d)



import datetime
import numpy
import typing
import uuid
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.utils import Geometry


class Graphic:

    def get_property(self, property: str):
        ...

    def mask_xdata_with_shape(self, shape: typing.Sequence[int]) -> DataAndMetadata.DataAndMetadata:
        """Return the mask created by this graphic as extended data.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def set_property(self, property: str, value) -> None:
        ...

    @property
    def angle(self) -> float:
        """Return the angle (radians) property."""
        ...

    @angle.setter
    def angle(self, value: float) -> None:
        """Set the angle (radians) property."""
        ...

    @property
    def bounds(self) -> typing.Tuple[typing.Tuple[float, float], typing.Tuple[float, float]]:
        """Return the bounds property in relative coordinates.

        Bounds is a tuple ((top, left), (height, width))"""
        ...

    @bounds.setter
    def bounds(self, value: typing.Tuple[typing.Tuple[float, float], typing.Tuple[float, float]]) -> None:
        """Set the bounds property in relative coordinates.

        Bounds is a tuple ((top, left), (height, width))"""
        ...

    @property
    def center(self) -> typing.Tuple[float, float]:
        """Return the center property in relative coordinates.

        Center is a tuple (y, x)."""
        ...

    @center.setter
    def center(self, value) -> None:
        """Set the center in relative coordinates.

        Center is a tuple (y, x)."""
        ...

    @property
    def end(self) -> typing.Union[float, typing.Tuple[float, float]]:
        """Return the end property in relative coordinates.

        End may be a float when graphic is an Interval or a tuple (y, x) when graphic is a Line."""
        ...

    @end.setter
    def end(self, value: typing.Union[float, typing.Tuple[float, float]]) -> None:
        """Set the end property in relative coordinates.

        End may be a float when graphic is an Interval or a tuple (y, x) when graphic is a Line."""
        ...

    @property
    def graphic_id(self) -> str:
        """Return the graphic identifier.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @graphic_id.setter
    def graphic_id(self, value: str) -> None:
        """Set the graphic identifier.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def graphic_type(self) -> str:
        """Return the type of this graphic.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def interval(self) -> typing.Tuple[float, float]:
        """Return the interval property in relative coordinates.

        Interval is a tuple of floats (start, end)."""
        ...

    @interval.setter
    def interval(self, value: typing.Tuple[float, float]) -> None:
        """Set the interval property in relative coordinates.

        Interval is a tuple of floats (start, end)."""
        ...

    @property
    def label(self) -> str:
        """Return the graphic label.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @label.setter
    def label(self, value: str) -> None:
        """Set the graphic label.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def position(self) -> typing.Tuple[float, float]:
        """Return the position property in relative coordinates.

        Position is a tuple of floats (y, x)."""
        ...

    @position.setter
    def position(self, value: typing.Tuple[float, float]) -> None:
        """Set the position property in relative coordinates.

        Position is a tuple of floats (y, x)."""
        ...

    @property
    def region(self) -> "Graphic":
        ...

    @property
    def size(self) -> typing.Tuple[float, float]:
        """Return the size property in relative coordinates.

        Size is a tuple of floats (height, width)."""
        ...

    @size.setter
    def size(self, value: typing.Tuple[float, float]) -> None:
        """Set the size property in relative coordinates.

        Size is a tuple of floats (height, width)."""
        ...

    @property
    def start(self) -> typing.Union[float, typing.Tuple[float, float]]:
        """Return the start property in relative coordinates.

        Start may be a float when graphic is an Interval or a tuple (y, x) when graphic is a Line."""
        ...

    @start.setter
    def start(self, value: typing.Union[float, typing.Tuple[float, float]]) -> None:
        """Set the end property in relative coordinates.

        End may be a float when graphic is an Interval or a tuple (y, x) when graphic is a Line."""
        ...

    @property
    def type(self) -> str:
        """Return the region type property.

        The region type is different from the preferred 'graphic_type' in that it is backwards compatible with older versions.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def uuid(self) -> uuid.UUID:
        """Return the uuid of this object.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def vector(self) -> typing.Tuple[typing.Tuple[float, float], typing.Tuple[float, float]]:
        """Return the vector property in relative coordinates.

        Vector will be a tuple of tuples ((y_start, x_start), (y_end, x_end))."""
        ...

    @vector.setter
    def vector(self, value: typing.Tuple[typing.Tuple[float, float], typing.Tuple[float, float]]) -> None:
        """Set the vector property in relative coordinates.

        Vector will be a tuple of tuples ((y_start, x_start), (y_end, x_end))."""
        ...


class DataItem:

    def add_channel_region(self, position: float) -> Graphic:
        ...

    def add_ellipse_region(self, center_y: float, center_x: float, height: float, width: float) -> Graphic:
        ...

    def add_interval_region(self, start: float, end: float) -> Graphic:
        ...

    def add_line_region(self, start_y: float, start_x: float, end_y: float, end_x: float) -> Graphic:
        ...

    def add_point_region(self, y: float, x: float) -> Graphic:
        """Add a point graphic to the data item.

        :param x: The x coordinate, in relative units [0.0, 1.0]
        :param y: The y coordinate, in relative units [0.0, 1.0]
        :return: The :py:class:`nion.swift.Facade.Graphic` object that was added.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def add_rectangle_region(self, center_y: float, center_x: float, height: float, width: float) -> Graphic:
        ...

    def delete_metadata_value(self, key: str) -> None:
        """Delete the metadata value for the given key.

        There are a set of predefined keys that, when used, will be type checked and be interoperable with other
        applications. Please consult reference documentation for valid keys.

        If using a custom key, we recommend structuring your keys in the '<dotted>.<group>.<attribute>' format followed
        by the predefined keys. e.g. 'stem.session.instrument' or 'stm.camera.binning'.

        Also note that some predefined keys map to the metadata ``dict`` but others do not. For this reason, prefer
        using the ``metadata_value`` methods over directly accessing ``metadata``.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def get_metadata_value(self, key: str) -> typing.Any:
        """Get the metadata value for the given key.

        There are a set of predefined keys that, when used, will be type checked and be interoperable with other
        applications. Please consult reference documentation for valid keys.

        If using a custom key, we recommend structuring your keys in the '<group>.<attribute>' format followed
        by the predefined keys. e.g. 'session.instrument' or 'camera.binning'.

        Also note that some predefined keys map to the metadata ``dict`` but others do not. For this reason, prefer
        using the ``metadata_value`` methods over directly accessing ``metadata``.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def has_metadata_value(self, key: str) -> bool:
        """Return whether the metadata value for the given key exists.

        There are a set of predefined keys that, when used, will be type checked and be interoperable with other
        applications. Please consult reference documentation for valid keys.

        If using a custom key, we recommend structuring your keys in the '<group>.<attribute>' format followed
        by the predefined keys. e.g. 'session.instrument' or 'camera.binning'.

        Also note that some predefined keys map to the metadata ``dict`` but others do not. For this reason, prefer
        using the ``metadata_value`` methods over directly accessing ``metadata``.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def mask_xdata(self) -> DataAndMetadata.DataAndMetadata:
        """Return the mask by combining any mask graphics on this data item as extended data.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def remove_region(self, graphic: Graphic) -> None:
        ...

    def set_data(self, data: numpy.ndarray) -> None:
        """Set the data.

        :param data: A numpy ndarray.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def set_data_and_metadata(self, data_and_metadata: DataAndMetadata.DataAndMetadata) -> None:
        """Set the data and metadata.

        :param data_and_metadata: The data and metadata.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def set_dimensional_calibrations(self, dimensional_calibrations: typing.List[Calibration.Calibration]) -> None:
        """Set the dimensional calibrations.

        :param dimensional_calibrations: A list of calibrations, must match the dimensions of the data.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def set_intensity_calibration(self, intensity_calibration: Calibration.Calibration) -> None:
        """Set the intensity calibration.

        :param intensity_calibration: The intensity calibration.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def set_metadata(self, metadata: dict) -> None:
        """Set the metadata dict.

        :param metadata: The metadata dict.

        The metadata dict must be convertible to JSON, e.g. ``json.dumps(metadata)`` must succeed.

        For best future compatibility, prefer using the ``get_metadata_value`` and ``set_metadata_value`` methods over
        directly accessing ``metadata``.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def set_metadata_value(self, key: str, value: typing.Any) -> None:
        """Set the metadata value for the given key.

        There are a set of predefined keys that, when used, will be type checked and be interoperable with other
        applications. Please consult reference documentation for valid keys.

        If using a custom key, we recommend structuring your keys in the '<group>.<attribute>' format followed
        by the predefined keys. e.g. 'session.instrument' or 'camera.binning'.

        Also note that some predefined keys map to the metadata ``dict`` but others do not. For this reason, prefer
        using the ``metadata_value`` methods over directly accessing ``metadata``.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def created(self) -> datetime.datetime:
        """Return the created timestamp (UTC) as a datetime object.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def data(self) -> numpy.ndarray:
        """Return the data as a numpy ndarray.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @data.setter
    def data(self, value: numpy.ndarray) -> None:
        """Set the data.

        :param data: A numpy ndarray.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def data_and_metadata(self) -> DataAndMetadata.DataAndMetadata:
        """Return the extended data.

        .. versionadded:: 1.0
        .. deprecated:: 1.1
           Use :py:attr:`~nion.swift.Facade.DataItem.xdata` instead.

        Scriptable: Yes
        """
        ...

    @property
    def dimensional_calibrations(self) -> typing.List[Calibration.Calibration]:
        """Return a copy of the list of dimensional calibrations.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def display(self) -> "Display":
        ...

    @property
    def display_xdata(self) -> DataAndMetadata.DataAndMetadata:
        """Return the extended data of this data item display.

        Display data will always be 1d or 2d and either int, float, or RGB data type.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def graphics(self) -> typing.List[Graphic]:
        """Return the graphics attached to this data item.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def intensity_calibration(self) -> Calibration.Calibration:
        """Return a copy of the intensity calibration.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def metadata(self) -> dict:
        """Return a copy of the metadata as a dict.

        For best future compatibility, prefer using the ``get_metadata_value`` and ``set_metadata_value`` methods over
        directly accessing ``metadata``.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def modified(self) -> datetime.datetime:
        """Return the modified timestamp (UTC) as a datetime object.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def regions(self) -> typing.List[Graphic]:
        """Return the graphics attached to this data item.

        .. versionadded:: 1.0
        .. deprecated:: 1.1
           Use :py:attr:`~nion.swift.Facade.DataItem.graphics` instead.

        Scriptable: Yes
        """
        ...

    @property
    def title(self) -> str:
        """Return the title as a string.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @title.setter
    def title(self, value: str) -> None:
        """Set the title to a string.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def uuid(self) -> uuid.UUID:
        """Return the uuid of this object.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def xdata(self) -> DataAndMetadata.DataAndMetadata:
        """Return the extended data of this data item.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @xdata.setter
    def xdata(self, value: DataAndMetadata.DataAndMetadata) -> None:
        """Set the extended data of this data item.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...


class DisplayPanel:

    def set_data_item(self, data_item: DataItem) -> None:
        """Set the data item associated with this display panel.

        :param data_item: The :py:class:`nion.swift.Facade.DataItem` object to add.

        This will replace whatever data item, browser, or controller is currently in the display panel with the single
        data item.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def data_item(self) -> DataItem:
        """Return the data item associated with this display panel.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...


class Display:

    def get_graphic_by_id(self, graphic_id: str) -> Graphic:
        ...

    @property
    def data_item(self) -> DataItem:
        ...

    @property
    def display_type(self) -> str:
        ...

    @display_type.setter
    def display_type(self, value: str) -> None:
        ...

    @property
    def graphics(self) -> typing.List[Graphic]:
        ...

    @property
    def selected_graphics(self) -> typing.List[Graphic]:
        ...

    @property
    def uuid(self) -> uuid.UUID:
        """Return the uuid of this object.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...


class DataGroup:

    def add_data_item(self, data_item: DataItem) -> None:
        """Add a data item to the group.

        :param data_item: The :py:class:`nion.swift.Facade.DataItem` object to add.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def uuid(self) -> uuid.UUID:
        """Return the uuid of this object.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...


class Library:

    def copy_data_item(self, data_item: DataItem) -> DataItem:
        """Copy a data item.

        .. versionadded:: 1.0

        Scriptable: No
        """
        ...

    def create_data_item(self, title: str=None) -> DataItem:
        """Create an empty data item in the library.

        :param title: The title of the data item (optional).
        :return: The new :py:class:`nion.swift.Facade.DataItem` object.
        :rtype: :py:class:`nion.swift.Facade.DataItem`

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def create_data_item_from_data(self, data: numpy.ndarray, title: str=None) -> DataItem:
        """Create a data item in the library from an ndarray.

        The data for the data item will be written to disk immediately and unloaded from memory. If you wish to delay
        writing to disk and keep using the data, create an empty data item and use the data item methods to modify
        the data.

        :param data: The data (ndarray).
        :param title: The title of the data item (optional).
        :return: The new :py:class:`nion.swift.Facade.DataItem` object.
        :rtype: :py:class:`nion.swift.Facade.DataItem`

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def create_data_item_from_data_and_metadata(self, data_and_metadata: DataAndMetadata.DataAndMetadata, title: str=None) -> DataItem:
        """Create a data item in the library from a data and metadata object.

        The data for the data item will be written to disk immediately and unloaded from memory. If you wish to delay
        writing to disk and keep using the data, create an empty data item and use the data item methods to modify
        the data.

        :param data_and_metadata: The data and metadata.
        :param title: The title of the data item (optional).
        :return: The new :py:class:`nion.swift.Facade.DataItem` object.
        :rtype: :py:class:`nion.swift.Facade.DataItem`

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def data_ref_for_data_item(self, data_item: DataItem):
        ...

    def delete_library_value(self, key: str) -> None:
        """Delete the library value for the given key.

        Please consult the developer documentation for a list of valid keys.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def get_data_item_by_uuid(self, data_item_uuid: uuid.UUID) -> DataItem:
        """Get the data item with the given UUID.

        .. versionadded:: 1.0

        Status: Provisional
        Scriptable: Yes
        """
        ...

    def get_data_item_for_hardware_source(self, hardware_source, channel_id: str=None, processor_id: str=None, create_if_needed: bool=False, large_format: bool=False) -> DataItem:
        """Get the data item associated with hardware source and (optional) channel id and processor_id. Optionally create if missing.

        :param hardware_source: The hardware_source.
        :param channel_id: The (optional) channel id.
        :param processor_id: The (optional) processor id for the channel.
        :param create_if_needed: Whether to create a new data item if none is found.
        :return: The associated data item. May be None.

        .. versionadded:: 1.0

        Status: Provisional
        Scriptable: Yes
        """
        ...

    def get_data_item_for_reference_key(self, data_item_reference_key: str=None, create_if_needed: bool=False, large_format: bool=False) -> DataItem:
        """Get the data item associated with data item reference key. Optionally create if missing.

        :param data_item_reference_key: The data item reference key.
        :param create_if_needed: Whether to create a new data item if none is found.
        :return: The associated data item. May be None.

        .. versionadded:: 1.0

        Status: Provisional
        Scriptable: Yes
        """
        ...

    def get_dependent_data_items(self, data_item: DataItem) -> typing.List[DataItem]:
        """Return the dependent data items the data item argument.

        :return: The list of :py:class:`nion.swift.Facade.DataItem` objects.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def get_graphic_by_uuid(self, graphic_uuid: uuid.UUID) -> Graphic:
        """Get the graphic with the given UUID.

        .. versionadded:: 1.0

        Status: Provisional
        Scriptable: Yes
        """
        ...

    def get_library_value(self, key: str) -> typing.Any:
        """Get the library value for the given key.

        Please consult the developer documentation for a list of valid keys.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def get_or_create_data_group(self, title: str) -> DataGroup:
        """Get (or create) a data group.

        :param title: The title of the data group.
        :return: The new :py:class:`nion.swift.Facade.DataGroup` object.
        :rtype: :py:class:`nion.swift.Facade.DataGroup`

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def get_source_data_items(self, data_item: DataItem) -> typing.List[DataItem]:
        """Return the list of data items that are data sources for the data item.

        :return: The list of :py:class:`nion.swift.Facade.DataItem` objects.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def has_library_value(self, key: str) -> bool:
        """Return whether the library value for the given key exists.

        Please consult the developer documentation for a list of valid keys.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def set_library_value(self, key: str, value: typing.Any) -> None:
        """Set the library value for the given key.

        Please consult the developer documentation for a list of valid keys.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def snapshot_data_item(self, data_item: DataItem) -> DataItem:
        """Snapshot a data item. Similar to copy but with a data snapshot.

        .. versionadded:: 1.0

        Scriptable: No
        """
        ...

    @property
    def data_item_count(self) -> int:
        """Return the data item count.

        :return: The number of data items.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def data_items(self) -> typing.List[DataItem]:
        """Return the list of data items.

        :return: The list of :py:class:`nion.swift.Facade.DataItem` objects.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def uuid(self) -> uuid.UUID:
        """Return the uuid of this object.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...


class DocumentWindow:

    def add_data(self, data: numpy.ndarray, title: str=None) -> DataItem:
        """Create a data item in the library from data.

        .. versionadded:: 1.0
        .. deprecated:: 1.1
           Use :py:meth:`~nion.swift.Facade.Library.create_data_item_from_data` instead.

        Scriptable: No
        """
        ...

    def create_data_item_from_data(self, data: numpy.ndarray, title: str=None) -> DataItem:
        """Create a data item in the library from data.

        .. versionadded:: 1.0
        .. deprecated:: 1.1
           Use :py:meth:`~nion.swift.Facade.Library.create_data_item_from_data` instead.

        Scriptable: No
        """
        ...

    def create_data_item_from_data_and_metadata(self, data_and_metadata: DataAndMetadata.DataAndMetadata, title: str=None) -> DataItem:
        """Create a data item in the library from the data and metadata.

        .. versionadded:: 1.0
        .. deprecated:: 1.1
           Use :py:meth:`~nion.swift.Facade.Library.create_data_item_from_data_and_metadata` instead.

        Scriptable: No
        """
        ...

    def display_data_item(self, data_item: DataItem, source_display_panel=None, source_data_item=None):
        """Display a new data item and gives it keyboard focus. Uses existing display if it is already displayed.

        .. versionadded:: 1.0

        Status: Provisional
        Scriptable: Yes
        """
        ...

    def get_display_panel_by_id(self, identifier: str) -> DisplayPanel:
        """Return display panel with the identifier.

        .. versionadded:: 1.0

        Status: Provisional
        Scriptable: Yes
        """
        ...

    def get_or_create_data_group(self, title: str) -> DataGroup:
        """Get (or create) a data group.

        .. versionadded:: 1.0
        .. deprecated:: 1.1
           Use :py:meth:`~nion.swift.Facade.Library.create_data_item_from_data` instead.

        Scriptable: No
        """
        ...

    def queue_task(self, fn) -> None:
        ...

    def show_confirmation_message_box(self, caption: str, accepted_fn, rejected_fn=None, accepted_text: str=None, rejected_text: str=None, display_rejected: str=False) -> None:
        ...

    def show_get_string_message_box(self, caption: str, text: str, accepted_fn, rejected_fn=None, accepted_text: str=None, rejected_text: str=None) -> None:
        """Show a dialog box and ask for a string.

        Caption describes the user prompt. Text is the initial/default string.

        Accepted function must be a function taking one argument which is the resulting text if the user accepts the
        message dialog. It will only be called if the user clicks OK.

        Rejected function can be a function taking no arguments, called if the user clicks Cancel.

        .. versionadded:: 1.0

        Scriptable: No
        """
        ...

    def show_modeless_dialog(self, item, handler=None):
        ...

    @property
    def all_display_panels(self) -> typing.List[DisplayPanel]:
        """Return the list of display panels currently visible.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def library(self) -> Library:
        """Return the library object.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def target_data_item(self) -> DataItem:
        ...

    @property
    def target_display(self) -> Display:
        ...


class Application:

    @property
    def document_controllers(self) -> typing.List[DocumentWindow]:
        """Return the document controllers.

        .. versionadded:: 1.0
        .. deprecated:: 1.1
           Use :py:attr:`~nion.swift.Facade.Application.document_windows` instead.

        Scriptable: Yes
        """
        ...

    @property
    def document_windows(self) -> typing.List[DocumentWindow]:
        """Return the document windows.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def library(self) -> Library:
        """Return the library object.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...


class API:
    """An interface to Nion Swift.

    This class cannot be instantiated directly. Use :samp:`api_broker.get_api(version)` to get access an instance of
    this class.
    """

    def create_calibration(self, offset: float=None, scale: float=None, units: str=None) -> Calibration.Calibration:
        """Create a calibration object with offset, scale, and units.

        :param offset: The offset of the calibration.
        :param scale: The scale of the calibration.
        :param units: The units of the calibration as a string.
        :return: The calibration object.

        .. versionadded:: 1.0

        Scriptable: Yes

        Calibrated units and uncalibrated units have the following relationship:
            :samp:`calibrated_value = offset + value * scale`
        """
        ...

    def create_data_and_metadata(self, data: numpy.ndarray, intensity_calibration: Calibration.Calibration=None, dimensional_calibrations: typing.List[Calibration.Calibration]=None, metadata: dict=None, timestamp: str=None, data_descriptor: DataAndMetadata.DataDescriptor=None) -> DataAndMetadata.DataAndMetadata:
        """Create a data_and_metadata object from data.

        :param data: an ndarray of data.
        :param intensity_calibration: An optional calibration object.
        :param dimensional_calibrations: An optional list of calibration objects.
        :param metadata: A dict of metadata.
        :param timestamp: A datetime object.
        :param data_descriptor: A data descriptor describing the dimensions.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def create_data_and_metadata_from_data(self, data: numpy.ndarray, intensity_calibration: Calibration.Calibration=None, dimensional_calibrations: typing.List[Calibration.Calibration]=None, metadata: dict=None, timestamp: str=None) -> DataAndMetadata.DataAndMetadata:
        """Create a data_and_metadata object from data.

        .. versionadded:: 1.0
        .. deprecated:: 1.1
           Use :py:meth:`~nion.swift.Facade.DataItem.create_data_and_metadata` instead.

        Scriptable: No
        """
        ...

    def create_data_and_metadata_io_handler(self, io_handler_delegate):
        """Create an I/O handler that reads and writes a single data_and_metadata.

        :param io_handler_delegate: A delegate object :py:class:`DataAndMetadataIOHandlerInterface`

        .. versionadded:: 1.0

        Scriptable: No
        """
        ...

    def create_data_descriptor(self, is_sequence: bool, collection_dimension_count: int, datum_dimension_count: int) -> DataAndMetadata.DataDescriptor:
        """Create a data descriptor.

        :param is_sequence: whether the descriptor describes a sequence of data.
        :param collection_dimension_count: the number of collection dimensions represented by the descriptor.
        :param datum_dimension_count: the number of datum dimensions represented by the descriptor.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def create_hardware_source(self, hardware_source_delegate):
        ...

    def create_menu_item(self, menu_item_handler):
        ...

    def create_panel(self, panel_delegate):
        """Create a utility panel that can be attached to a window.

        .. versionadded:: 1.0

        Scriptable: No

         The panel_delegate should respond to the following:
            (property, read-only) panel_id
            (property, read-only) panel_name
            (property, read-only) panel_positions (a list from "top", "bottom", "left", "right", "all")
            (property, read-only) panel_position (from "top", "bottom", "left", "right", "none")
            (method, required) create_panel_widget(ui), returns a widget
            (method, optional) close()
        """
        ...

    def get_all_hardware_source_ids(self) -> typing.List[str]:
        ...

    def get_all_instrument_ids(self) -> typing.List[str]:
        ...

    def get_hardware_source_by_id(self, hardware_source_id: str, version: str):
        """Return the hardware source API matching the hardware_source_id and version.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    def get_instrument_by_id(self, instrument_id: str, version: str):
        ...

    def queue_task(self, fn) -> None:
        ...

    @property
    def application(self) -> Application:
        """Return the application object.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

    @property
    def library(self) -> Library:
        """Return the library object.

        .. versionadded:: 1.0

        Scriptable: Yes
        """
        ...

version = "~1.0"
